/*

Copyright (C) 2017 Pantxo Diribarne

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Octave is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include <iostream>

#include <QApplication>
#include <QtCore>
#include <QDebug>

#include <QFontDatabase>
#include <QImage>
#include <QPrinter>
#include <QPixmap>
#include <QPainter>
#include <QRegExp>
#include <QtXml>

// Painter classes
class basepainter : public QPainter
{
public:
  basepainter (void) :
  m_fname(""), m_sizef(), m_dpi(72.0)
  { }

  basepainter (QString fname, QRectF sizepix, double dpi, int reconstruct) :
    m_fname(fname), m_sizef(sizepix), m_dpi(dpi), m_reconstruct(reconstruct)
  {
    double scl = get_scale ();
    m_sizef.setWidth (m_sizef.width () * scl);
    m_sizef.setHeight (m_sizef.height () * scl);
  }

  ~basepainter (void) { }

  bool do_reconstruct (void) const { return m_reconstruct != 0; }

  QString get_fname (void) const { return m_fname; }

  QRectF get_rectf (void) const { return m_sizef; }

  QRect get_rect (void) const { return m_sizef.toRect (); }

  double get_scale (void) const { return m_dpi / 72.0; }

  virtual void finish (void) = 0;

private:
  QString m_fname;
  QRectF m_sizef;
  double m_dpi;
  int m_reconstruct;
};

class pdfpainter : public basepainter
{
public:
  pdfpainter (QString fname, QRectF sizepix, double dpi, int reconstruct) :
    basepainter (fname, sizepix, dpi, reconstruct), m_printer()
  {
    // Printer settings
    m_printer.setOutputFormat (QPrinter::PdfFormat);
    m_printer.setFontEmbeddingEnabled (true);
    m_printer.setOutputFileName (get_fname ());
    m_printer.setFullPage (true);
    m_printer.setPaperSize (get_rectf ().size (), QPrinter::DevicePixel);

    // Painter settings 
    begin (&m_printer);
    setViewport (get_rect ());
    scale (get_scale (), get_scale ());
  }

  ~pdfpainter (void) { }

  void finish (void) { end (); }

private:
  QPrinter m_printer;
};

class rasterpainter : public basepainter
{
public:
  rasterpainter (QString fname, QRectF sizepix, double dpi, int reconstruct, 
                 const char *fmt = 0):
    basepainter (fname, sizepix, dpi, reconstruct), m_image (), m_fmt(fmt)
  {
    // Scale
    QRect size = get_rect ();
    m_image = QPixmap (size.size ());

    // Painter settings
    begin (& m_image);
    setViewport (size);
    scale (get_scale (), get_scale ());
  }

  ~rasterpainter (void) { }

  void finish (void)
  {
    end ();
    m_image.save (get_fname (), m_fmt);
  }

private:
  QPixmap m_image;
  const char *m_fmt;
};


// String conversion functions
QVector<double> qstr2vectorf (QString str)
{
  QVector<double> pts;
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 1)
    {
      double pt = (*p).toDouble ();
      pts.append (pt);
    }
  return pts;
}

QVector<double> qstr2vectord (QString str)
{
  QVector<double> pts;
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 1)
    {
      double pt = (*p).toDouble ();
      pts.append (pt);
    }
  return pts;
}

QVector<QPointF> qstr2ptsvector (QString str)
{
  QVector<QPointF> pts;
  str.replace (" ", ",");
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 2)
    {
      QPointF pt ((*p).toDouble (), (*(p+1)).toDouble ());
      pts.append (pt);
    }
  return pts;
}

QVector<QPoint> qstr2ptsvectord (QString str)
{
  QVector<QPoint> pts;
  str.replace (" ", ",");
  QStringList coords = str.split (",");
  for (QStringList::iterator p = coords.begin (); p != coords.end (); p += 2)
    {
      QPoint pt ((*p).toDouble (), (*(p+1)).toDouble ());
      pts.append (pt);
    }
  return pts;
}

// Extract field arguments in a style-like string, e.g. "bla field(1,34,56) bla"

QString get_field (QString str, QString field)
{
  QString retval;
  QRegExp rx (field + "\\(([^\\)]*)\\)");
  int pos = 0;
  pos = rx.indexIn (str, pos);
  if (pos > -1)
    retval = rx.cap (1);
  return retval;
}

// Polygon reconstruction class
class octave_polygon
{
public:
  octave_polygon (void) :
    m_polygon(), m_color(QColor ("black"))
  { }

  octave_polygon (QPolygonF p, QColor c) :
    m_polygon(p), m_color(c)
  { }

  ~octave_polygon (void) { }

  void draw (basepainter &painter, bool do_clear = false)
  {
    if (m_polygon.count () > 2)
      {
        painter.setBrush (m_color);
        painter.setPen (Qt::NoPen);

        painter.setRenderHint (QPainter::Antialiasing, false);
        painter.drawPolygon (m_polygon);
        painter.setRenderHint (QPainter::Antialiasing, true);
        if (do_clear)
          m_polygon.clear ();
      }
  }

  bool merge (QPolygonF p, QColor c)
  {
    if (m_polygon.count () == 0)
      {
        m_polygon = p;
        m_color = c;
        return true;
      }

    if (m_color != c)
      return false;

    bool retval = false;
    QPolygonF tmp = m_polygon.united (p);
    if (tmp.count () <= (m_polygon.count () + 30))
      {
        m_polygon = tmp;
        retval = true;
      }
    else
      qDebug () << "Counts: " << tmp.count ()
                << ".vs." << m_polygon.count () << "\n";

    return retval;
  }

private:
  QPolygonF m_polygon;
  QColor m_color;
};

void draw (QDomElement& parent_elt, basepainter& painter)
{
  QDomNodeList nodes = parent_elt.childNodes ();

  static QString clippath_id;
  static QMap< QString, QVector<QPoint> > clippath;

  static octave_polygon current_polygon;

  // tspan elements must have access to the font and position extracted from
  // their parent text element
  static QFont font;
  static double dx = 0, dy = 0;

  for (int i = 0; i < nodes.count (); i++)
    {
      QDomNode node = nodes.at (i);
      if (! node.isElement ())
        return; //FIXME: continue?

      QDomElement elt = node.toElement ();

      // Draw eventual remaining polygon
      if (elt.tagName () != "polygon" && painter.do_reconstruct ())
        current_polygon.draw (painter, true);

      if (elt.tagName () == "clipPath")
        {
          clippath_id = "#" + elt.attribute ("id");
          draw (elt, painter);
          clippath_id = QString ();
        }
      else if (elt.tagName () == "g")
        {
          bool current_clipstate = painter.hasClipping ();
          QRegion current_clippath = painter.clipRegion ();

          QString str = elt.attribute ("clip-path");
          if (! str.isEmpty ())
            {
              QVector<QPoint> pts = clippath[get_field (str, "url")];
              if (! pts.isEmpty ())
                {
                  painter.setClipRegion (QRegion (QPolygon (pts)));
                  painter.setClipping (true);
                }
            }

          draw (elt, painter);

          // Draw eventual remaining polygon
          if (painter.do_reconstruct ())
            current_polygon.draw (painter, true);

          // Restore previous clipping settings
          painter.setClipRegion (current_clippath);
          painter.setClipping (current_clipstate);
        }
      else if (elt.tagName () == "text")
        {
          // Font
          font = QFont ();
          QString str = elt.attribute ("font-family");
          if (! str.isEmpty ())
            font.setFamily (elt.attribute ("font-family"));

          str = elt.attribute ("font-weight");
          if (! str.isEmpty () && str != "normal")
            font.setWeight (QFont::Bold);
          
          int sz = elt.attribute ("font-size").toInt ();
          if (sz > 0)
            font.setPixelSize (sz);

          painter.setFont (font);

          // Translation and rotation
          painter.save ();
          str = get_field (elt.attribute ("transform"), "translate");
          if (! str.isEmpty ())
            {
              QStringList trans = str.split (",");
              dx = trans[0].toDouble ();
              dy = trans[1].toDouble ();

              str = get_field (elt.attribute ("transform"), "rotate");
              if (! str.isEmpty ())
                {
                  QStringList rot = str.split (",");
                  painter.translate (dx+rot[1].toDouble (),
                                     dy+rot[2].toDouble ());
                  painter.rotate (rot[0].toDouble ());
                  dx = rot[1].toDouble ();
                  dy = rot[2].toDouble ();
                }
              else
                {
                  painter.translate (dx, dy);
                  dx = 0;
                  dy = 0;
                }
            }

          draw (elt, painter);
          painter.restore ();
        }
      else if (elt.tagName () == "tspan")
        {
          // Font
          QFont saved_font(font);

          QString str = elt.attribute ("font-family");
          if (! str.isEmpty ())
            font.setFamily (elt.attribute ("font-family"));

          str = elt.attribute ("font-weight");
          if (! str.isEmpty ())
            {
              if (str != "normal")
                font.setWeight (QFont::Bold);
              else
                font.setWeight (QFont::Normal);
            }

          int sz = elt.attribute ("font-size").toInt ();
          if (sz > 0)
            font.setPixelSize (sz);

          painter.setFont (font);

          // Color is specified in rgb
          str = get_field (elt.attribute ("fill"), "rgb");
          if (! str.isEmpty ())
            {
              QStringList clist = str.split (",");
              painter.setPen (QColor (clist[0].toInt (), clist[1].toInt (),
                                      clist[2].toInt ()));
            }

          QStringList xx = elt.attribute ("x").split (" ");
          int y = elt.attribute ("y").toInt ();
          str = elt.text ();
          if (! str.isEmpty ())
            {
              int ii = 0;
              foreach (QString s,  xx)
                if (ii < str.size ())
                  painter.drawText (s.toInt ()-dx, y-dy, str.at (ii++));
            }

          draw (elt, painter);
          font = saved_font;
        }
      else if (elt.tagName () == "polyline")
        {
          // Color
          QColor c (elt.attribute ("stroke"));
          QString str = elt.attribute ("stroke-opacity");
          if (! str.isEmpty () && str.toDouble () != 1.0
              && str.toDouble () >= 0.0)
            c.setAlphaF (str.toDouble ());

          QPen pen;
          pen.setColor (c);

          // Line properies
          str = elt.attribute ("stroke-width");
          if (! str.isEmpty ())
            {
              double w = str.toDouble () * painter.get_scale ();
              if (w > 0)
                pen.setWidthF (w / painter.get_scale ());
            }

          str = elt.attribute ("stroke-linecap");
          pen.setCapStyle (Qt::SquareCap);
          if (str == "round")
            pen.setCapStyle (Qt::RoundCap);
          else if (str == "butt")
            pen.setCapStyle (Qt::FlatCap);

          str = elt.attribute ("stroke-linejoin");
          pen.setJoinStyle (Qt::MiterJoin);
          if (str == "round")
            pen.setJoinStyle (Qt::RoundJoin);
          else if (str == "bevel")
            pen.setJoinStyle (Qt::BevelJoin);

          str = elt.attribute ("stroke-dasharray");
          pen.setStyle (Qt::SolidLine);
          if (! str.isEmpty ())
            {
              QVector<double> pat = qstr2vectord (str);
              if (pat.count () != 2 || pat[1] != 0)
                pen.setDashPattern (qstr2vectord (str));
            }

          painter.setPen (pen);
          painter.drawPolyline (qstr2ptsvector (elt.attribute ("points")));
        }
      else if (elt.tagName () == "image")
        {
          // Images are represented as a base64 stream of png formated data
          QString href_att = elt.attribute ("xlink:href");
          QString prefix ("data:image/png;base64,");
          QByteArray data =
            QByteArray::fromBase64(href_att.mid (prefix.length ())
                                   .toLocal8Bit ());
          QImage img;
          if (img.loadFromData (data, "PNG"))
            {
              QRect pos(elt.attribute ("x").toInt (),
                        elt.attribute ("y").toInt (),
                        elt.attribute ("width").toInt (),
                        elt.attribute ("height").toInt ());

              // Translate
              painter.save ();
              QString str = get_field (elt.attribute ("transform"), "matrix");
              if (! str.isEmpty ())
                {
                  QVector<double> m = qstr2vectorf (str);
                  double scl = painter.get_scale ();
                  QTransform tform(m[0]*scl, m[1]*scl, m[2]*scl,
                                   m[3]*scl, m[4]*scl, m[5]*scl);
                  painter.setTransform (tform);
                }

              painter.drawImage (pos, img);
              painter.restore  ();
            }
        }
      else if (elt.tagName () == "polygon")
        {
          if (! clippath_id.isEmpty ())
            clippath[clippath_id] = qstr2ptsvectord (elt.attribute ("points"));
          else
            {
              QString str = elt.attribute ("fill");
              if (! str.isEmpty ())
                {
                  QColor color (str);

                  QString str = elt.attribute ("fill-opacity");
                  if (! str.isEmpty () && str.toDouble () != 1.0
                      && str.toDouble () >= 0.0)
                    color.setAlphaF (str.toDouble ());

                  QPolygonF p (qstr2ptsvector (elt.attribute ("points")));
                  if (! painter.do_reconstruct ())
                    {
                      current_polygon = octave_polygon (p, color);
                      current_polygon.draw (painter);                      
                    }
                  else if (! current_polygon.merge (p, color))
                    {
                      // Draw previous polygon
                      current_polygon.draw (painter);
                      current_polygon = octave_polygon (p, color);
                    }
                }
            }
        }
    }
}

QTextStream& outstream ()
{
  static QTextStream ts (stdout);
  return ts;
}

int main(int argc, char *argv[])
{
  const char *help =
"Usage:\n\
\tqt-svg-convert infile fmt dpi font reconstruct outfile\n\n\
Convert svg file to pdf, or raster file formats. All arguments are mandatory:\n\
* infile: input svg file or \"-\"to indicate that the input svg file should be \
read from stdin\n\
* fmt: format of the output file. May be one of pdf, bmp, jpg, png or ppm\n\
* dpi: device dependent resolution in screen pixel per inch\n\
* font: specify a file name for the default FreeSans font\n\
* reconstruct: specify wether to reconstruct polygons (0 or 1)\n\
* outfile: output file name\n";

  if (argc != 7)
    {
      printf ("%s", help);
      return -1;
    }

  // Create a DOM document
  QDomDocument document;

  // Open a file for reading
  QFile file;
  if (! strcmp (argv[1], "-"))
    {
      // Read from stdin
      if (! file.open (stdin, QIODevice::ReadOnly | QIODevice::Text))
        {
          std::cout << "Unable read from stdin" << std::endl;
          printf ("%s", help);
          return -1;
        }
    }
  else
    {
      file.setFileName (argv[1]);
      if(! file.open (QIODevice::ReadOnly | QIODevice::Text))
        {
          std::cout << "Unable to open file " << argv[1] << std::endl;
          printf ("%s", help);
          return -1;
        }
    }

  // Load xml file
  QString msg;
  if(! document.setContent (&file, false, &msg))
    {
      std::cout << "Failed to load the file for reading: "
                << msg.toStdString ();
      printf ("%s", help);
      file.close();
      return -1;
    }
  
  // Format and resolution
  if (strcmp (argv[2], "pdf") && strcmp (argv[2], "bmp")
      && strcmp (argv[2], "jpg") && strcmp (argv[2], "png")
      && strcmp (argv[2], "ppm"))
    {
      printf ("Unhandled output file format %s\n%s", argv[2], help);
      return -1;
    }

  double dpi = QString (argv[3]).toDouble ();
  if (dpi <= 0.0)
    {
      printf ("dpi must be positive");
      return -1;
    }

  QString fontname (argv[4]);

  int reconstruct = QString (argv[5]).toInt ();

  // Get viewport from the root element
  QDomElement root = document.firstChildElement();
  double x0, y0, dx, dy;
  QString s = root.attribute ("viewBox");
  QTextStream (&s) >> x0 >> y0 >> dx >> dy;

  QRectF vp (x0, y0, dx, dy);

  // Setup application and add defult FreeSans font if needed
  QApplication a(argc, argv);

  QFont font ("FreeSans");
  int id = 0;
  if (! font.exactMatch ())
    {
      if (! fontname.isEmpty ())
        {
          id = QFontDatabase::addApplicationFont (fontname);
          if (id < 0)
            qDebug ("Unable to add default font to database\n");
        }
      else
        qDebug ("FreeSans font not found\n");
    }

  QTemporaryFile fout;
  if (! fout.open ())
    {
      printf ("Could not open temporary file\n");
      return -1;
    }

  if (! strcmp (argv[2], "pdf"))
    {
      // PDF painter
      pdfpainter painter (fout.fileName (), vp, dpi, reconstruct);
      draw (root, painter);
      painter.finish ();
    }
  else
    {
      // Raster painter
      rasterpainter rpainter (fout.fileName (), vp, dpi, reconstruct, argv[2]);
      draw (root, rpainter);
      rpainter.finish ();
    }
  
  if (QFile::exists (argv[6]))
    if (! QFile::remove (argv[6]))
      {
        qDebug () << "Unable to replace existing file " << argv[6] << "\n";
        return -1;
      }

  fout.copy (argv[6]);

  return 0;
}
